package com.fjhx.flow.service.flow.impl;

import cn.hutool.core.util.StrUtil;
import cn.hutool.extra.spring.SpringUtil;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.fjhx.flow.core.FlowBean;
import com.fjhx.flow.core.FlowDelegate;
import com.fjhx.flow.core.FlowThreadLocalUtil;
import com.fjhx.flow.entity.flow.dto.InitiateDto;
import com.fjhx.flow.entity.flow.dto.JumpDto;
import com.fjhx.flow.entity.flow.po.FlowDefinition;
import com.fjhx.flow.entity.flow.po.FlowDefinitionNode;
import com.fjhx.flow.entity.flow.po.FlowExample;
import com.fjhx.flow.entity.flow.po.FlowExampleDetail;
import com.fjhx.flow.enums.FlowStatusEnum;
import com.fjhx.flow.enums.HandleTypeEnum;
import com.fjhx.flow.enums.NodeTypeEnum;
import com.fjhx.flow.service.flow.*;
import com.googlecode.aviator.AviatorEvaluator;
import com.googlecode.aviator.Expression;
import com.ruoyi.common.constant.StatusConstant;
import com.ruoyi.common.core.domain.BaseIdPo;
import com.ruoyi.common.exception.ServiceException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.TransactionStatus;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;

@Slf4j
@Service
public class FlowProcessServiceImpl implements FlowProcessService {

    @Autowired
    private FlowDefinitionService flowDefinitionService;

    @Autowired
    private FlowDefinitionNodeService flowDefinitionNodeService;

    @Autowired
    private FlowExampleService flowExampleService;

    @Autowired
    private FlowExampleDetailService flowExampleDetailService;

    @Autowired
    private PlatformTransactionManager platformTransactionManager;

    @Autowired
    private TransactionDefinition transactionDefinition;

    @Override
    public void initiate(InitiateDto dto) {

        // 找到代码定义的流程bean
        Class<? extends FlowDelegate> flowDelegateCls = FlowBean.getBean(dto.getFlowKey());

        // 获取流程委托对象
        FlowDelegate flowDelegate = getFlowDelegate(flowDelegateCls);

        // 查找可用流程
        FlowDefinition flowDefinition = flowDefinitionService.getOne(q -> q
                .eq(FlowDefinition::getFlowKey, dto.getFlowKey())
                .eq(FlowDefinition::getCurrentVersion, StatusConstant.YES));

        if (flowDefinition == null) {
            throw new ServiceException("没有找到可用流程");
        }

        // 流程节点列表
        List<FlowDefinitionNode> flowDefinitionNodeList = flowDefinitionNodeService.list(
                q -> q.eq(FlowDefinitionNode::getFlowDefinitionId, flowDefinition.getId()));

        // 父级节点map
        Map<Long, List<FlowDefinitionNode>> parentNodeMap = flowDefinitionNodeList.stream()
                .collect(Collectors.groupingBy(FlowDefinitionNode::getParentId));

        // 寻找开始节点
        FlowDefinitionNode startNode = flowDefinitionNodeList.stream()
                .filter(item -> NodeTypeEnum.START.getKey().equals(item.getNodeType()))
                .findFirst().orElse(null);

        if (startNode == null) {
            throw new ServiceException("没有找到开始节点");
        }

        // 生成流程id
        long flowId = IdWorker.getId();

        // 初始化模板参数
        Map<String, Object> templateMap = flowDelegate.initTemplateMap(dto.getData(), dto.getData());

        FlowThreadLocalUtil.setTemplateData(templateMap);
        FlowThreadLocalUtil.setStartData(dto.getData());
        FlowThreadLocalUtil.setCurrentData(dto.getData());
        FlowThreadLocalUtil.setFlowId(flowId);

        // 寻找下一节点
        FlowDefinitionNode nextUserNode = getNextUserNode(startNode, parentNodeMap);

        // 流程实例
        FlowExample flowExample = new FlowExample();

        // 流程实例明细列表
        List<FlowExampleDetail> flowExampleDetailList = new ArrayList<>();

        // 执行开始流程方法
        Long businessId = flowDelegate.start(flowId, dto.getData());
        FlowThreadLocalUtil.setBusinessId(businessId);

        FlowExampleDetail startExampleDetail = new FlowExampleDetail();
        startExampleDetail.setFlowExampleId(flowId);
        startExampleDetail.setFlowDefinitionNodeId(startNode.getId());
        startExampleDetail.setFlowDefinitionNodeType(startNode.getNodeType());
        startExampleDetail.setHandleType(HandleTypeEnum.NEXT.getKey());
        startExampleDetail.setHandleRemark(dto.getRemark());
        flowExampleDetailList.add(startExampleDetail);

        // 如果下个节点是结束节点
        if (NodeTypeEnum.END.equals(NodeTypeEnum.getEnum(nextUserNode.getNodeType()))) {

            // 添加明细
            invokeEndMethod(nextUserNode, flowDelegateCls, flowDelegate);

            // 结束流程
            FlowExampleDetail endExampleDetail = new FlowExampleDetail();
            endExampleDetail.setFlowExampleId(flowId);
            endExampleDetail.setFlowDefinitionNodeId(nextUserNode.getId());
            endExampleDetail.setFlowDefinitionNodeType(nextUserNode.getNodeType());
            endExampleDetail.setHandleType(HandleTypeEnum.OVER.getKey());
            flowExampleDetailList.add(endExampleDetail);

            // 流程已通过
            flowExample.setStatus(FlowStatusEnum.HAVE_PASSED.getKey());
        } else {
            // 流程进行中
            flowExample.setStatus(FlowStatusEnum.IN_PROGRESS.getKey());
        }

        flowExample.setTitle(StrUtil.format(flowDefinition.getTitleTemplate(), templateMap, true));
        flowExample.setFlowKey(dto.getFlowKey());
        flowExample.setDefinitionId(flowDefinition.getId());
        flowExample.setDefinitionNodeId(nextUserNode.getId());
        flowExample.setBusinessId(businessId);
        flowExample.setStartData(dto.getData().toJSONString());

        // 开启一个事务
        TransactionStatus transaction = platformTransactionManager.getTransaction(transactionDefinition);

        try {
            flowExampleService.save(flowExample);
            flowExampleDetailService.saveBatch(flowExampleDetailList);

            // 提交事务
            platformTransactionManager.commit(transaction);
        } catch (Exception e) {
            // 回滚事务
            platformTransactionManager.rollback(transaction);
            log.error("保存流程异常", e);
            throw new ServiceException("保存流程异常");
        }

    }

    @Override
    public void jump(JumpDto dto) {
        Long flowId = dto.getFlowId();

        FlowExample flowExample = flowExampleService.getById(flowId);
        if (flowExample == null) {
            throw new ServiceException("没有找到流程");
        }

        if (!flowExample.getVersion().equals(dto.getVersion())) {
            throw new ServiceException("流程已被处理");
        }

        Integer status = flowExample.getStatus();
        if (status > 1) {
            throw new ServiceException("流程已结束");
        }

        // 找到代码定义的流程bean
        Class<? extends FlowDelegate> flowDelegateCls = FlowBean.getBean(flowExample.getFlowKey());

        // 获取流程委托对象
        FlowDelegate flowDelegate = getFlowDelegate(flowDelegateCls);

        // 流程节点列表
        List<FlowDefinitionNode> flowDefinitionNodeList = flowDefinitionNodeService.list(
                q -> q.eq(FlowDefinitionNode::getFlowDefinitionId, flowExample.getDefinitionId()));

        // 获取当前审批节点
        Long definitionNodeId = flowExample.getDefinitionNodeId();
        FlowDefinitionNode currentNode = flowDefinitionNodeList.stream()
                .filter(item -> item.getId().equals(definitionNodeId)).findFirst().orElse(null);

        if (currentNode == null) {
            throw new ServiceException("没有找到当前审批节点");
        }

        // 查找跳转类型
        HandleTypeEnum handleTypeEnum = HandleTypeEnum.getEnum(dto.getHandleType());

        // 流程实例明细列表
        List<FlowExampleDetail> flowExampleDetailList = new ArrayList<>();

        // 节点处理
        FlowExampleDetail nodeExampleDetail = new FlowExampleDetail();
        nodeExampleDetail.setFlowExampleId(flowId);
        nodeExampleDetail.setFlowDefinitionNodeId(currentNode.getId());
        nodeExampleDetail.setFlowDefinitionNodeType(currentNode.getNodeType());
        nodeExampleDetail.setHandleType(dto.getHandleType());
        nodeExampleDetail.setHandleRemark(dto.getRemark());
        nodeExampleDetail.setSubmitData(dto.getData().toJSONString());
        flowExampleDetailList.add(nodeExampleDetail);

        FlowDefinitionNode jumpUserNode = null;

        // 发起流程参数
        JSONObject startDataJson = JSONObject.parseObject(flowExample.getStartData());
        Map<String, Object> templateMap = flowDelegate.initTemplateMap(dto.getData(), startDataJson);
        FlowThreadLocalUtil.setTemplateData(templateMap);
        FlowThreadLocalUtil.setStartData(dto.getData());
        FlowThreadLocalUtil.setCurrentData(dto.getData());
        FlowThreadLocalUtil.setFlowId(flowId);
        FlowThreadLocalUtil.setBusinessId(flowExample.getBusinessId());
        FlowThreadLocalUtil.setHandleTypeEnum(handleTypeEnum);

        FlowStatusEnum flowStatusEnum;

        switch (handleTypeEnum) {
            case NEXT:
                // 父级节点map
                Map<Long, List<FlowDefinitionNode>> parentNodeMap = flowDefinitionNodeList.stream()
                        .collect(Collectors.groupingBy(FlowDefinitionNode::getParentId));
                // 寻找下一节点
                jumpUserNode = getNextUserNode(currentNode, parentNodeMap);
                // 赋值流程所在id
                flowExample.setDefinitionNodeId(jumpUserNode.getId());
                // 如果下个节点是结束节点
                if (NodeTypeEnum.END.equals(NodeTypeEnum.getEnum(jumpUserNode.getNodeType()))) {
                    // 流程已通过
                    flowStatusEnum = FlowStatusEnum.HAVE_PASSED;

                    // 结束流程
                    FlowExampleDetail endExampleDetail = new FlowExampleDetail();
                    endExampleDetail.setFlowExampleId(flowId);
                    endExampleDetail.setFlowDefinitionNodeId(jumpUserNode.getId());
                    endExampleDetail.setFlowDefinitionNodeType(jumpUserNode.getNodeType());
                    endExampleDetail.setHandleType(HandleTypeEnum.OVER.getKey());
                    flowExampleDetailList.add(endExampleDetail);
                } else {
                    // 流程进行中
                    flowStatusEnum = FlowStatusEnum.IN_PROGRESS;
                }
                break;

            case PREVIOUS:
                // 节点id对象map
                Map<Long, FlowDefinitionNode> nodeMap = flowDefinitionNodeList.stream()
                        .collect(Collectors.toMap(BaseIdPo::getId, Function.identity()));
                // 查找上一个节点
                jumpUserNode = getLastOneUserNode(currentNode, nodeMap);
                // 赋值流程所在id
                flowExample.setDefinitionNodeId(jumpUserNode.getId());
                // 如果流程回退到开始节点
                if (NodeTypeEnum.START.equals(NodeTypeEnum.getEnum(jumpUserNode.getNodeType()))) {
                    // 流程进行中
                    flowStatusEnum = FlowStatusEnum.UNINITIATED;
                } else {
                    // 流程进行中
                    flowStatusEnum = FlowStatusEnum.IN_PROGRESS;
                }
                break;

            case OVER:
                flowStatusEnum = FlowStatusEnum.REJECTED;
                break;

            default:
                throw new ServiceException("未知节点处理类型");
        }

        flowExample.setStatus(flowStatusEnum.getKey());

        // 执行当前节点方法
        String handlingMethod = currentNode.getHandlingMethod();
        if (StrUtil.isNotBlank(handlingMethod)) {
            invokeMethod(flowDelegateCls, flowDelegate, handlingMethod);
        }

        // 执行结束方法
        if (jumpUserNode != null && NodeTypeEnum.END.equals(NodeTypeEnum.getEnum(jumpUserNode.getNodeType()))) {
            invokeEndMethod(jumpUserNode, flowDelegateCls, flowDelegate);
        }

        // 驳回通用方法
        if (handleTypeEnum.equals(HandleTypeEnum.OVER)) {
            flowDelegate.rejected();
        }

        // 开启一个事务
        TransactionStatus transaction = platformTransactionManager.getTransaction(transactionDefinition);

        try {
            flowExampleService.updateById(flowExample);
            flowExampleDetailService.saveBatch(flowExampleDetailList);
            // 提交事务
            platformTransactionManager.commit(transaction);
        } catch (Exception e) {
            // 提交事务
            platformTransactionManager.rollback(transaction);
            log.error("保存流程异常", e);
            throw new ServiceException("保存流程异常");
        }

    }

    /**
     * 执行结束方法
     *
     * @param nextUserNode    下个用户节点
     * @param flowDelegateCls 流程代理class
     * @param flowDelegate    流程代理对象
     */
    private void invokeEndMethod(FlowDefinitionNode nextUserNode, Class<? extends FlowDelegate> flowDelegateCls, FlowDelegate flowDelegate) {

        // 走非默认方法
        String handlingMethod = nextUserNode.getHandlingMethod();
        if (StrUtil.isNotBlank(handlingMethod)) {
            // 执行节点方法
            invokeMethod(flowDelegateCls, flowDelegate, handlingMethod);
        } else {
            // 执行流程结束方法
            flowDelegate.end();
        }

    }

    /**
     * 执行节点方法
     *
     * @param flowDelegateCls 流程代理class
     * @param flowDelegate    流程代理对象
     * @param handlingMethod  流程执行方法名
     */
    private void invokeMethod(Class<? extends FlowDelegate> flowDelegateCls, FlowDelegate flowDelegate, String handlingMethod) {
        try {
            Method method = flowDelegateCls.getMethod(handlingMethod.trim());
            method.invoke(flowDelegate);
        } catch (Exception e) {
            log.error("自定义处理方法异常:{}", handlingMethod, e);
            throw new ServiceException("自定义处理方法异常：" + handlingMethod);
        }
    }

    /**
     * 获取流程委托对象
     *
     * @param flowDelegateCls 流程代理class
     * @return 流程委托对象
     */
    private FlowDelegate getFlowDelegate(Class<? extends FlowDelegate> flowDelegateCls) {
        FlowDelegate flowDelegate = SpringUtil.getBean(flowDelegateCls);
        if (flowDelegate == null) {
            throw new ServiceException("没有把流程bean加入到spring中");
        }
        return flowDelegate;
    }

    /**
     * 查找下一个用户执行节点
     *
     * @param currentNode   当前节点
     * @param parentNodeMap 父节点map
     * @return 用户节点
     */
    private FlowDefinitionNode getNextUserNode(FlowDefinitionNode currentNode, Map<Long, List<FlowDefinitionNode>> parentNodeMap) {
        // 查找下个节点
        List<FlowDefinitionNode> nextNodeList = parentNodeMap.get(currentNode.getId());
        FlowDefinitionNode nextNode = null;
        // 如果为分支
        if (NodeTypeEnum.BRANCH.equals(NodeTypeEnum.getEnum(currentNode.getNodeType()))) {
            // 循环分支
            for (FlowDefinitionNode flowDefinitionNode : nextNodeList) {
                // 获取跳转条件
                String jumpCondition = flowDefinitionNode.getJumpCondition();
                // 跳转条件为空表示默认跳转
                if (StrUtil.isBlank(jumpCondition)) {
                    nextNode = flowDefinitionNode;
                }
                // 符合条件
                if (expressionResult(FlowThreadLocalUtil.getTemplateData(), jumpCondition)) {
                    nextNode = flowDefinitionNode;
                    break;
                }
            }
            if (nextNode == null) {
                throw new ServiceException("未识别到分支跳转节点");
            }
        }
        // 如果不为分支
        else {
            if (nextNodeList == null || nextNodeList.size() != 1) {
                throw new ServiceException("流程定义错误");
            }
            // 查找下一个节点方法
            nextNode = nextNodeList.get(0);
        }
        // 如果下一节点为分支，递归找下一节点
        if (NodeTypeEnum.BRANCH.equals(NodeTypeEnum.getEnum(nextNode.getNodeType()))) {
            return getNextUserNode(nextNode, parentNodeMap);
        }
        return nextNode;
    }

    /**
     * 查找上一个节点
     *
     * @param currentNode 当前节点
     * @param nodeMap     id节点map
     * @return 用户节点
     */
    private FlowDefinitionNode getLastOneUserNode(FlowDefinitionNode currentNode, Map<Long, FlowDefinitionNode> nodeMap) {
        FlowDefinitionNode lastOneUserNode = nodeMap.get(currentNode.getParentId());
        if (NodeTypeEnum.BRANCH.equals(NodeTypeEnum.getEnum(lastOneUserNode.getNodeType()))) {
            return getLastOneUserNode(lastOneUserNode, nodeMap);
        }
        return lastOneUserNode;
    }

    /**
     * 校验el表达式
     */
    public boolean expressionResult(Map<String, Object> map, String expression) {
        Expression exp = AviatorEvaluator.compile(expression);
        Object execute = exp.execute(map);
        return Boolean.parseBoolean(String.valueOf(execute));
    }

}
